; File: ir-rx-pulse-dist.asm

; == Possible values of register ir_pulseCntr ==
; Pulse/pause # within a frame, incremented at each falling edge (= begin of pulse):
; 0 = no edge received yet, waiting for begin of start pulse; or pulse already too long -> to be discarded
; 1 = after start edge, start pulse or pause ongoing
; 2 = first data pulse or pause ongoing, after reception of start pulse+pause
; 3 = second data pulse or pause ongoing, after reception of first data pulse+pause
; .
; .
; .
; 33 = last data pulse or pause ongoing
; 34 = stop pulse ongoing
.equ IR_DATA_START = 3 ; value of ir_pulseCntr when first data pulse+pause has been received
.equ IR_REPETITION_PAUSE_RECEIVED = 200 ; = repetition pause received, possible repetition frame stop bit ongoing

; == Debug message codes ==
.equ IR_PULSE_TOO_SHORT = $01
.equ IR_PULSE_TOO_LONG = $02
.equ IR_PULSE_TIMEOUT = $03
.equ IR_PAUSE_TOO_SHORT = $04
.equ IR_PAUSE_BETWEEN_0_1 = $05
.equ IR_PAUSE_TOO_LONG = $06
.equ IR_PAUSE_START_TOO_SHORT = $07
.equ IR_PAUSE_START_BETWEEN_R_D = $08
.equ IR_PAUSE_START_TOO_LONG = $09
.equ IR_PULSE_START_TOO_SHORT = $0A
.equ IR_PULSE_START_TOO_LONG = $0B
.equ IR_PULSE_COUNT_INVALID = $10
.equ IR_DATA_INVALID = $20

; == IR code definitions ==
; Time values must be in ascending order, otherwise the corresponding comparisons must be modified.
.if IR_PROTOCOL == IR_NEC
.equ IR_PULSE = 560 ; µs
.equ IR_PAUSE_0 = 560 ; µs
.equ IR_PAUSE_1 = 1690 ; µs
.equ IR_PAUSE_START_REPETITION = 2250 ; µs
.equ IR_PAUSE_START_DATA = 4500 ; µs
.equ IR_PULSE_START = 9000 ; µs
.equ IR_PULSE_COUNT = 34 ; including start and stop bit
.elif IR_PROTOCOL == IR_SAMSUNG32
.equ IR_PULSE = 550 ; µs
.equ IR_PAUSE_0 = 550 ; µs
.equ IR_PAUSE_1 = 1650 ; µs
.equ IR_PAUSE_START_DATA = 4500 ; µs
.equ IR_PULSE_START = 4500 ; µs
.equ IR_PULSE_COUNT = 34 ; including start and stop bit
.endif

; == Calculations ==
; IR_DELAY = delay between IR interrupts in µs
#define IR_DELAY (1000000.0/IR_INTR_FREQ)
#define IR_TIMEOUT_INTR ROUND(1000.0*(IR_TIMEOUT)/IR_DELAY)
#define IR_TIMER_OCR ROUND(1.0*F_CPU/IR_TIMER_PRESC/IR_INTR_FREQ-1)
; Calculate the approximate interrupt counts of time constants (T is time in µs):
#define IR_MIN_INTR(T) ROUND((1.0-IR_TOLERANCE)*(T)/IR_DELAY)
#define IR_MAX_INTR(T) ROUND((1.0+IR_TOLERANCE)*(T)/IR_DELAY)+1

; == Routines ==
; === IR decoder interrupt routine ===
ir_intr:
; Save temporary registers and SREG:
	push	temp0
	push	temp1
	xin		temp0, SREG
	push	temp0
; Increment interrupt counter:
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
	subi	temp0, low(-1) ; increment temp1:temp0 ...
	sbci	temp1, high(-1) ; ... 16-bit
	mov		ir_intrCntrL, temp0
	mov		ir_intrCntrH, temp1
; Check whether timeout is reached:
	subi	temp0, low(IR_TIMEOUT_INTR)
	sbci	temp1, high(IR_TIMEOUT_INTR)
	brsh	ir_intr_timeout
; Detect whether IR_RX value has changed:
	xin		temp0, IR_RX_PIN
	mov		temp1, temp0							; save current IR_RX_PIN value
	eor		temp0, ir_prev
	sbrs	temp0, IR_RX
	rjmp	ir_intr_end								; no change
; IR_RX has changed -> either falling or rising edge.
	mov		ir_prev, temp1							; update previous value
	sbrc	temp1, IR_RX
	rjmp	ir_intr_risingEdge						; rising edge
; Falling edge.
	rcall	ir_fallingEdge
	rjmp	ir_intr_end
ir_intr_risingEdge:
	rcall	ir_risingEdge
	rjmp	ir_intr_end
ir_intr_timeout:
	rcall	ir_timeout
ir_intr_end:
; Restore modified registers:
	pop		temp0
	xout	SREG, temp0
	pop		temp1
	pop		temp0
	reti

; === Create (1<<temp0) bit mask in temp1 ===
; Destroys temp0
lslTemp1Temp0:
	andi	temp0, 0x07
	ldi		temp1, 1
	tst		temp0
lslTemp1Temp0_loop:
	breq	lslTemp1Temp0_end
	lsl		temp1
	dec		temp0
	rjmp	lslTemp1Temp0_loop
lslTemp1Temp0_end:
	ret

; === Set IR data bit ===
ir_setDataBit:
	push	temp0
	push	temp1
	mov		temp0, ir_pulseCntr
	subi	temp0, IR_DATA_START
	cpi		temp0, 8
	brlo	ir_setDataBit_0
	cpi		temp0, 16
	brlo	ir_setDataBit_1
	cpi		temp0, 24
	brlo	ir_setDataBit_2
; temp0 >= 24
	rcall	lslTemp1Temp0
	or		ir_data3, temp1
	rjmp	ir_setDataBit_end
ir_setDataBit_2:
; 16 <= temp0 < 24
	rcall	lslTemp1Temp0
	or		ir_data2, temp1
	rjmp	ir_setDataBit_end
ir_setDataBit_1:
; 8 <= temp0 < 16
	rcall	lslTemp1Temp0
	or		ir_data1, temp1
	rjmp	ir_setDataBit_end
ir_setDataBit_0:
; 0 <= temp0 < 8
	rcall	lslTemp1Temp0
	or		ir_data0, temp1
ir_setDataBit_end:
	pop		temp1
	pop		temp0
	ret

; === IR falling edge handler ===
ir_fallingEdge:
	push	temp0
	push	temp1
	push	temp2
	inc		ir_pulseCntr
	cpi		ir_pulseCntr, 1
	breq	ir_fallingEdge_end					; first falling edge -- done
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
	cpi		ir_pulseCntr, 2
	breq	ir_fallingEdge_2nd					; second falling edge (after start pause)
; Falling edge after data pause.
.ifdef IR_DEBUG
	ldi		temp2, IR_PAUSE_TOO_SHORT
.endif
	subi	temp0, low(IR_MIN_INTR(IR_PAUSE_0))
	sbci	temp1, high(IR_MIN_INTR(IR_PAUSE_0))
	brlo	ir_fallingEdge_discard				; pause was too short
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
	subi	temp0, low(IR_MAX_INTR(IR_PAUSE_0))
	sbci	temp1, high(IR_MAX_INTR(IR_PAUSE_0))
	brlo	ir_fallingEdge_end					; '0' received -- done
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
.ifdef IR_DEBUG
	ldi		temp2, IR_PAUSE_BETWEEN_0_1
.endif
	subi	temp0, low(IR_MIN_INTR(IR_PAUSE_1))
	sbci	temp1, high(IR_MIN_INTR(IR_PAUSE_1))
	brlo	ir_fallingEdge_discard
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
	subi	temp0, low(IR_MAX_INTR(IR_PAUSE_1))
	sbci	temp1, high(IR_MAX_INTR(IR_PAUSE_1))
	brlo	ir_fallingEdge_setDataBit			; '1' received
.ifdef IR_DEBUG
	ldi		temp2, IR_PAUSE_TOO_LONG
.endif
	rjmp	ir_fallingEdge_discard				; pause was too long
ir_fallingEdge_2nd:
.ifdef IR_DEBUG
	ldi		temp2, IR_PAUSE_START_TOO_SHORT
.endif
.ifdef IR_RECOGNIZE_REPETITION
	subi	temp0, low(IR_MIN_INTR(IR_PAUSE_START_REPETITION))
	sbci	temp1, high(IR_MIN_INTR(IR_PAUSE_START_REPETITION))
	brlo	ir_fallingEdge_discard
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
	subi	temp0, low(IR_MAX_INTR(IR_PAUSE_START_REPETITION))
	sbci	temp1, high(IR_MAX_INTR(IR_PAUSE_START_REPETITION))
	brlo	ir_fallingEdge_afterRepPause
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
.ifdef IR_DEBUG
	ldi		temp2, IR_PAUSE_START_BETWEEN_R_D
.endif
.endif
	subi	temp0, low(IR_MIN_INTR(IR_PAUSE_START_DATA))
	sbci	temp1, high(IR_MIN_INTR(IR_PAUSE_START_DATA))
	brlo	ir_fallingEdge_discard
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
.ifdef IR_DEBUG
	ldi		temp2, IR_PAUSE_START_TOO_LONG
.endif
	subi	temp0, low(IR_MAX_INTR(IR_PAUSE_START_DATA))
	sbci	temp1, high(IR_MAX_INTR(IR_PAUSE_START_DATA))
	brsh	ir_fallingEdge_discard
; Start pause was within bounds.
	clr		ir_data0
	clr		ir_data1
	clr		ir_data2
	clr		ir_data3
	rjmp	ir_fallingEdge_end
ir_fallingEdge_discard:
.ifdef IR_DEBUG
	rcall	ir_discardHandler
.endif
	clr		ir_pulseCntr
	rjmp	ir_fallingEdge_end
ir_fallingEdge_setDataBit:
	rcall	ir_setDataBit
.ifdef IR_RECOGNIZE_REPETITION
	rjmp	ir_fallingEdge_end
ir_fallingEdge_afterRepPause:
	ldi		ir_pulseCntr, IR_REPETITION_PAUSE_RECEIVED
.endif
ir_fallingEdge_end:
	clr		ir_intrCntrL
	clr		ir_intrCntrH
.if IR_TIMER_PRESC > 1
; Reset Timer2 prescaler, ATmega88A
	ldi		temp0, 1<<PSRASY
	xout	GTCCR, temp0
.endif
; Restore registers:
	pop		temp2
	pop		temp1
	pop		temp0
	ret

; === IR rising edge handler ===
ir_risingEdge:
	push	temp0
	push	temp1
	push	temp2
	cpi		ir_pulseCntr, 0
	breq	ir_risingEdge_end					; do nothing if timeout occurred during previous pulse
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
	cpi		ir_pulseCntr, 1
	breq	ir_risingEdge_1st
; Data or stop rising edge.
.ifdef IR_DEBUG
	ldi		temp2, IR_PULSE_TOO_SHORT
.endif
	subi	temp0, low(IR_MIN_INTR(IR_PULSE))
	sbci	temp1, high(IR_MIN_INTR(IR_PULSE))
	brlo	ir_risingEdge_discard				; pulse was too short
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
.ifdef IR_DEBUG
	ldi		temp2, IR_PULSE_TOO_LONG
.endif
	subi	temp0, low(IR_MAX_INTR(IR_PULSE))
	sbci	temp1, high(IR_MAX_INTR(IR_PULSE))
	brsh	ir_risingEdge_discard				; pulse was too long
; Data or stop pulse was within bounds.
	rjmp	ir_risingEdge_end
ir_risingEdge_1st:
.ifdef IR_DEBUG
	ldi		temp2, IR_PULSE_START_TOO_SHORT
.endif
	subi	temp0, low(IR_MIN_INTR(IR_PULSE_START))
	sbci	temp1, high(IR_MIN_INTR(IR_PULSE_START))
	brlo	ir_risingEdge_discard				; start pulse was too short
	mov		temp0, ir_intrCntrL
	mov		temp1, ir_intrCntrH
.ifdef IR_DEBUG
	ldi		temp2, IR_PULSE_START_TOO_LONG
.endif
	subi	temp0, low(IR_MAX_INTR(IR_PULSE_START))
	sbci	temp1, high(IR_MAX_INTR(IR_PULSE_START))
	brsh	ir_risingEdge_discard				; start pulse was too long
; Start pulse was within bounds.
	rjmp	ir_risingEdge_end
ir_risingEdge_discard:
.ifdef IR_DEBUG
	rcall	ir_discardHandler
.endif
	clr		ir_pulseCntr
ir_risingEdge_end:
	clr		ir_intrCntrL
	clr		ir_intrCntrH
.if IR_TIMER_PRESC > 1
; Reset Timer2 prescaler, ATmega88A
	ldi		temp0, 1<<PSRASY
	xout	GTCCR, temp0
.endif
; Restore registers:
	pop		temp2
	pop		temp1
	pop		temp0
	ret

; === IR timeout handler ===
ir_timeout:
	push	temp2
.ifdef IR_DEBUG
	ldi		temp2, IR_PULSE_TIMEOUT
.endif
	sbrs	ir_prev, IR_RX							; If timeout occurred during pulse ...
	rjmp	ir_timeout_discard						; ... discard current reception.
; Timeout occurred during pause.
	tst		ir_pulseCntr
	breq	ir_timeout_end
.ifdef IR_RECOGNIZE_REPETITION
	cpi		ir_pulseCntr, IR_REPETITION_PAUSE_RECEIVED
	breq	ir_timeout_repetition
.endif
.ifdef IR_DEBUG
	ldi		temp2, IR_PULSE_COUNT_INVALID
.endif
	cpi		ir_pulseCntr, IR_PULSE_COUNT
	brne	ir_timeout_discard
; Complete data frame received.
.if IR_PROTOCOL == IR_NEC
; Check integrity of address and command:
	com		ir_data1
	com		ir_data3
.ifdef IR_DEBUG
	ldi		temp2, IR_DATA_INVALID
.endif
	cp		ir_data0, ir_data1
	brne	ir_timeout_discard
	cp		ir_data2, ir_data3
	brne	ir_timeout_discard
.endif
	rcall	ir_validDataHandler
	rjmp	ir_timeout_clr
ir_timeout_discard:
.ifdef IR_DEBUG
	rcall	ir_discardHandler
.endif
	rjmp	ir_timeout_clr
.ifdef IR_RECOGNIZE_REPETITION
ir_timeout_repetition:
	rcall	ir_repetitionHandler
.endif
ir_timeout_clr:
	clr		ir_pulseCntr
ir_timeout_end:
	clr		ir_intrCntrL
	clr		ir_intrCntrH
	pop		temp2
	ret
