; File:		ir-rx-pulse-dist-mega88a\main.asm
; Created:	18 Apr 2023 16:42:18
; Version:	2023-05-09
; Author:	Johannes Fechner
;			https://www.mikrocontroller.net/user/show/jofe
; Device:	ATmega88A

.include <mega88Adef.inc>

; Set the CPU clock frequency in Hz:
.equ F_CPU = 7_372_800
; Set the UART baud rate in bits per second:
.equ UART_BPS = 9_600

.def zero = r2
; Temporary registers
.def temp0 = r16
.def temp1 = r17
.def temp2 = r18
; IR registers
.def ir_prev = r3
.def ir_intrcntrl = r4
.def ir_intrcntrh = r5
.def ir_pulsecntr = r19
.def ir_data0 = r6
.def ir_data1 = r7
.def ir_data2 = r8
.def ir_data3 = r9

.org 0
	rjmp	reset
.org OC2Aaddr
	rjmp	intr_ir									; Timer2 Compare Match A interrupt
.org INT_VECTORS_SIZE

.include "macros.asm"
.include "ir-rx-pulse-dist.config.asm"
.include "ir-rx-pulse-dist.asm"

reset:
; initialize stack pointer
	ldi		temp0, high(RAMEND)
	xout	SPH, temp0
	ldi		temp0, low(RAMEND)
	xout	SPL, temp0
; initialize registers
	clr		zero
	ldi		temp0, 1<<IR_RX							; low-active
	mov		ir_prev, temp0
	clr		ir_intrcntrl
	clr		ir_intrcntrh
	clr		ir_pulsecntr
; initialize UART: 8 data bits, no parity, 1 stop bit
	ldi		temp0, high(ROUND(F_CPU/(16*UART_BPS)-1))
	xout	UBRR0H, temp0
	ldi		temp0, low(ROUND(F_CPU/(16*UART_BPS)-1))
	xout	UBRR0L, temp0
	ldi		temp0, (1<<UCSZ01) | (1<<UCSZ00)
	xout	UCSR0C, temp0
	ldi		temp0, 1<<TXEN0							; transmitting only
	xout	UCSR0B, temp0
; initialize Timer2
	ldi		temp0, IR_TIMER_OCR
	xout	OCR2A, temp0
	ldi		temp0, 1<<WGM21							; CTC mode
	xout	TCCR2A, temp0
	ldi		temp0, 1<<CS21							; prescaler 1/8
	xout	TCCR2B, temp0
	ldi		temp0, 1<<OCIE2A						; enable Compare Match A interrupt
	xout	TIMSK2, temp0
	sei
; main loop
loop:
	rjmp	loop

; === Routine: send character via UART ===
; Parameters: temp0 = the character to be transmitted
rtn_uart_send_ch:
	push	temp0
rtn_uart_send_ch_wait:
	xin		temp0, UCSR0A
	sbrs	temp0, UDRE0
	rjmp	rtn_uart_send_ch_wait
	pop		temp0
	xout	UDR0, temp0
	ret

; === Routine: send string via UART ===
; ZH:ZL must contain the start address of the string which has to be null-terminated.
rtn_uart_send_sz:
	push	temp0
rtn_uart_send_sz_loop:
	lpm		temp0, Z+
	tst		temp0
	breq	rtn_uart_send_sz_end
	rcall	rtn_uart_send_ch
	rjmp	rtn_uart_send_sz_loop
rtn_uart_send_sz_end:
	pop		temp0
	ret

; === Routine: send number in hexadecimal via UART ===
; Eine 8-Bit-Zahl ohne Vorzeichen hexadezimal ausgeben
; Übergabe: Byte im Register temp0
rtn_uart_send_hex:
	push	temp0
	swap	temp0
	andi	temp0, 0x0F
	rcall	rtn_uart_send_hex_digit
	pop		temp0
	push	temp0
	andi	temp0, 0x0F
	rcall	rtn_uart_send_hex_digit
	pop		temp0
	ret
rtn_uart_send_hex_digit:
	cpi		temp0, 10
	brlt	rtn_uart_send_hex_digit_1
	subi	temp0, -('A'-'9'-1)
rtn_uart_send_hex_digit_1:
	subi	temp0, -'0'
	rcall	rtn_uart_send_ch
	ret

; === Routine: handle discarded IR reception ===
.ifdef IR_DEBUG
rtn_ir_discard_handler:
	ldiz	2*sz_discarded
	rcall	rtn_uart_send_sz
	mov		temp0, temp2
	rcall	rtn_uart_send_hex
	ldiz	2*sz_pulsecntr
	rcall	rtn_uart_send_sz
	mov		temp0, ir_pulsecntr
	rcall	rtn_uart_send_hex
	ldiz	2*sz_intcntr
	rcall	rtn_uart_send_sz
	mov		temp0, ir_intrcntrh
	rcall	rtn_uart_send_hex
	mov		temp0, ir_intrcntrl
	rcall	rtn_uart_send_hex
	ldiz	2*sz_newline
	rcall	rtn_uart_send_sz
	ret
.endif

; === Routine: handle IR reception of valid data frame ===
rtn_ir_valid_handler:
	ldiz	2*sz_address
	rcall	rtn_uart_send_sz
.if IR_PROTOCOL == IR_SAMSUNG32
	mov		temp0, ir_data1
	rcall	rtn_uart_send_hex
.endif
	mov		temp0, ir_data0
	rcall	rtn_uart_send_hex
	ldiz	2*sz_command
	rcall	rtn_uart_send_sz
.if IR_PROTOCOL == IR_SAMSUNG32
	mov		temp0, ir_data3
	rcall	rtn_uart_send_hex
.endif
	mov		temp0, ir_data2
	rcall	rtn_uart_send_hex
	ldiz	2*sz_newline
	rcall	rtn_uart_send_sz
	ret

; === Routine: handle IR reception of repetition frame ===
.ifdef PAUSE_START_REPETITION
rtn_ir_repetition_handler:
	ldiz	2*sz_repetition
	rcall	rtn_uart_send_sz
	ldiz	2*sz_newline
	rcall	rtn_uart_send_sz
	ret
.endif

; === Data output strings ===
sz_address:
.db "Valid reception; address: $", 0
sz_command:
.db "; command: $", 0, 0
.ifdef PAUSE_START_REPETITION
sz_repetition:
.db "Valid reception; repetition frame", 0
.endif
sz_newline:
.db '\r', '\n', 0, 0

; === Debug message strings ===
.ifdef IR_DEBUG
sz_discarded:
.db "Discarded reception; code: $", 0, 0
sz_intcntr:
.db "; intcntr: $", 0, 0
sz_pulsecntr:
.db "; pulsecntr: $", 0, 0
.endif
